# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""LDAP Distinguished Name (DN) utilities."""

from __future__ import annotations

import functools
from typing import Self

import ldap.dn

from freeiam.errors import InvalidDN
from freeiam.ldap.constants import AVA, DNFormat


__all__ = ('DN',)


@functools.lru_cache
def _to_dn(dn):
    return ldap.dn.str2dn(dn)


class DN:
    """A LDAP Distinguished Name."""

    _CASE_INSENSITIVE_ATTRIBUTES = ('c', 'cn', 'dc', 'l', 'o', 'ou', 'uid')

    __slots__ = ('_cached_hash', '_cached_normalized', '_dn', '_format', 'dn')

    @classmethod
    def get(cls, dn: Self | str) -> Self:
        """Get a DN from string or existing DN."""
        return cls(dn) if isinstance(dn, str) else dn

    @classmethod
    def escape(cls, value: str) -> str:
        """Escape LDAP DN value."""
        return ldap.dn.escape_dn_chars(value)

    @classmethod
    def compose(cls, *parts: DN | str | tuple[str, str] | tuple[str, str, int]) -> Self:
        """
        Compose a DN from different segments.

        >>> base = DN('dc=freeiam,dc=org')
        >>> str(DN.compose(('cn', 'admin'), 'ou=foo,ou=bar', base))
        "cn=admin,ou=foo,ou=bar,dc=freeiam,dc=org"
        """
        rdns = []
        for part in parts:
            if isinstance(part, DN):
                rdns.extend(part.rdns)
            elif isinstance(part, str):
                rdns.extend(cls(part).rdns)
            elif isinstance(part, tuple) and len(part) > 1:
                rdns.append([[*part[:3], AVA.String][:3]])
            else:
                raise TypeError(part)
        return cls(ldap.dn.dn2str(rdns))

    @classmethod
    def normalize(cls, dn: Self | str) -> str:
        """Normalize DN."""
        return str(cls.get(dn))

    @classmethod
    def get_unique(cls, dns: list[str]) -> set[Self]:
        """
        Return a unique set of DNs.

        >>> len(DN.unique(['CN=users,dc=freeiam,dc=org', 'cn=users,dc=freeiam,dc=org', 'cn = users,dc=freeiam,dc=org', 'CN=Users,dc=freeiam,dc=org']))
        1
        """
        return {cls(dn) for dn in dns}

    @classmethod
    def get_unique_str(cls, dns: list[Self]) -> set[str]:
        """
        Return a unique set of DN strings from DNs.

        >>> DN.get_unique_str(DN.unique(['cn=foo', 'cn=bar']) - DN.unique(['cn = foo'])) == {'cn=bar'}
        True
        """
        return {str(dn) for dn in dns}

    @property
    def rdn(self) -> tuple[str, str]:
        """
        Get attr and value of the first RDN component.

        >>> DN('cn=foo,cn=bar').rdn
        ('cn', 'foo')
        """
        try:
            return self.multi_rdn[0]
        except IndexError:
            return ()

    @property
    def attribute(self) -> str:
        """
        Get attribute name of the first RDN component.

        >>> DN('cn=foo,cn=bar').attribute
        'cn'
        """
        try:
            return self.rdn[0]
        except IndexError:
            return ()

    @property
    def value(self) -> str:
        """
        Get value of the first RDN component.

        >>> DN('cn=foo,cn=bar').value
        'foo'
        """
        try:
            return self.rdn[1]
        except IndexError:
            return ()

    @property
    def multi_rdn(self) -> tuple[tuple[str, str]]:
        """
        Get all attrs and values of the RDN.

        >>> DN('uid=1+cn=2,dc=3').rdn
        (('uid', '1'), ('cn', '2'))
        """
        try:
            return tuple(tuple(rdn[:2]) for rdn in self._dn[0])
        except IndexError:
            return ()

    @property
    def attributes(self) -> tuple[str]:
        """
        Get attribute name of the first RDN component.

        >>> DN('uid=1+cn=2,dc=3').attributes
        ('uid', 'cn')
        """
        try:
            return tuple(rdn[0] for rdn in self._dn[0])
        except IndexError:
            return ()

    @property
    def values(self) -> tuple[str]:
        """
        Get value of the first RDN component.

        >>> DN('uid=1+cn=2,dc=3').values
        ('1', '2')
        """
        try:
            return tuple(rdn[1] for rdn in self._dn[0])
        except IndexError:
            return ()

    @property
    def rdns(self) -> list[list[tuple[str, str, int]]]:
        """Get the single RDN items."""
        return self._dn

    @property
    def parent(self) -> Self | None:
        """
        Get the parent DN.

        >>> DN('cn=item,cn=parent').parent == DN('cn=parent')
        True
        """
        if len(self._dn) > 1:
            return self[1:]
        return None

    def __init__(self, dn: str, format: DNFormat = None) -> None:  # noqa: A002
        self.dn = dn
        self._format = format
        self._cached_hash = None
        self._cached_normalized = None
        try:
            self._dn = _to_dn(self.dn)
        except ldap.DECODING_ERROR:
            try:
                self._dn = _to_dn(self.dn.replace(r'\?', '?'))  # Samba LDAP returns broken DN
            except ldap.DECODING_ERROR as exc:
                err = InvalidDN()
                err._description = 'Malformed DN syntax'
                err._info = f'{self.dn!r}: {exc}'.removesuffix(': ')
                raise err from exc

    def get_parent(self, end: Self | str) -> Self | None:
        """
        Get the parent DN until a certain base.

        >>> base = DN('dc=freeiam,dc=org')
        >>> DN('cn=foo,dc=freeiam,dc=org').get_parent(base) == base
        True
        >>> DN('dc=freeiam,dc=org').get_parent(base)
        None
        """
        if not self.endswith(end) or self == end:
            return None
        return self.parent

    def endswith(self, other: Self | str) -> bool:
        """
        Check if DN is descendant of another base DN.

        >>> DN('cn=foo,cn=bar').endswith('cn=bar')
        True
        >>> DN('cn=foo,cn=bar').endswith('cn=foo')
        False
        >>> DN('cn=foo').endswith('cn=foo,cn=bar')
        False
        >>> DN('cn=foo,cn=bar').endswith('')
        True
        """
        other = self.get(other)
        return self[-len(other) or len(self) :] == other

    def startswith(self, other: Self | str):
        """
        Check if DN starts with another DN.

        >>> DN('cn=foo,cn=bar').startswith('cn=foo')
        True
        >>> DN('cn=foo,cn=bar').startswith('cn=bar')
        False
        >>> DN('cn=foo,cn=bar').startswith('')
        True
        """
        other = self.get(other)
        return self[: len(other)] == other

    def walk(self, base: Self | str | None = None):
        """
        Walk the reversed DN components from the given base.

        >>> [str(x) for x in DN('cn=foo,cn=bar,cn=baz,cn=blub').walk('cn=baz,cn=blub')]
        ['cn=baz,cn=blub', 'cn=bar,cn=baz,cn=blub', 'cn=foo,cn=bar,cn=baz,cn=blub']
        >>> [str(x) for x in DN('cn=foo,cn=bar,cn=baz,cn=blub').walk()]
        ['cn=blub', 'cn=baz,cn=blub', 'cn=bar,cn=baz,cn=blub', 'cn=foo,cn=bar,cn=baz,cn=blub']
        """
        base = self.get(base or '')
        if not self.endswith(base):
            msg = 'DN does not end with given base'
            raise ValueError(msg)

        for i in reversed(range(len(self) - (len(base) or 1) + 1)):
            yield self[i:]

    def __str__(self) -> str:
        """
        Get a normalized string representation of the DN.

        >>> str(DN('cn = foo , cn = bar')) == "cn=foo,cn=bar"
        True
        """
        if self._cached_normalized is None:
            self._cached_normalized = ldap.dn.dn2str(self._dn)
        return self._cached_normalized

    def __repr__(self) -> str:
        """
        Get a representation.

        >>> repr(DN('cn=foo,cn=bar')) == "DN('cn=foo,cn=bar')"
        True
        """
        return f'{type(self).__name__}({str(self)!r})'

    def __len__(self) -> int:
        """Return number of components of the DN."""
        return len(self._dn)

    def __getitem__(self, key: str | slice) -> Self:
        """Get slice or item of the DN components."""
        if isinstance(key, slice):
            return self.__class__(ldap.dn.dn2str(self._dn[key]))
        return self.__class__(ldap.dn.dn2str([self._dn[key]]))

    def __eq__(self, other: Self | str) -> bool:
        """
        Check normalized DNs for equality.

        >>> DN('cn=foo') == DN('cn=foo')
        True
        >>> DN('cn=foo') == DN('cn=bar')
        False
        >>> DN('Cn=Foo') == DN('cn=foo')
        True
        >>> DN('Cn=foo') == DN('cn=bar')
        False
        >>> DN('uid=Administrator') == DN('uid=administrator')
        True
        >>> DN('foo=Foo') == DN('foo=foo')
        False
        >>> DN('cn=foo,cn=bar') == DN('cn=foo,cn=bar')
        True
        >>> DN('cn=bar,cn=foo') == DN('cn=foo,cn=bar')
        False
        >>> DN('cn=foo+cn=bar') == DN('cn=foo+cn=bar')
        True
        >>> DN('cn=bar+cn=foo') == DN('cn=foo+cn=bar')
        True
        >>> DN('cn=bar+Cn=foo') == DN('cn=foo+Cn=bar')
        True
        >>> DN(r'cn=%s31foo' % chr(92)) == DN(r'cn=1foo')
        True
        """
        return hash(self) == hash(DN(other) if isinstance(other, str) else other)

    def __ne__(self, other: Self | str) -> bool:
        return not self == other

    def __hash__(self) -> str:
        if self._cached_hash is None:
            self._cached_hash = hash(tuple(
                tuple(sorted(
                    (attr.lower(), val.lower() if attr.lower() in self._CASE_INSENSITIVE_ATTRIBUTES else val, ava)
                    for attr, val, ava in rdn
                )) for rdn in self._dn
            ))  # fmt: skip
        return self._cached_hash

    def __add__(self, other):
        return self.__class__(f'{self},{other}')



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""LDAP filter expressions."""

import enum
import operator
import re
import string
from collections import deque
from collections.abc import Callable, Sequence
from typing import Self

import lark
import ldap.filter
from lark import Lark, Transformer, v_args

from freeiam import errors


__all__ = (
    'AND',
    'NOT',
    'OR',
    'ApproximateMatch',
    'Comparison',
    'EqualityMatch',
    'ExtensibleMatch',
    'Filter',
    'GreaterOrEqual',
    'LessOrEqual',
    'Operator',
    'PresenceMatch',
    'SubstringMatch',
)


LDAP_FILTER_GRAMMAR = r"""
start: _group | bare
bare: comparison
_group: ws? (operator | expression) ws?

?operator: and | or | not

expression: "(" ows comparison ows ")"
and: "(" ows "&" ows groups ows ")"
or:  "(" ows "|" ows groups ows ")"
not: "(" ows "!" ows _group  ows ")"
?groups: _group* -> groups

?comparison: attr "=*"                         -> presence
     | attr "="  value                          -> equality
     | attr "="  substrings                     -> substring
     | attr ">=" value                          -> ge
     | attr "<=" value                          -> le
     | attr "~=" value                          -> approx
     |      ":"  dn ":" matchingrule ":=" value -> extmatch_noattr_dn_match
     | attr ":"  dn ":" matchingrule ":=" value -> extmatch_attr_dn_match
     | attr ":"  dn                  ":=" value -> extmatch_attr_dn_nomatch
     |      ":"         matchingrule ":=" value -> extmatch_noattr_nodn_match
     | attr ":"         matchingrule ":=" value -> extmatch_attr_nodn_match
     | attr                          ":=" value -> extmatch_attr_nodn_nomatch

substrings: substr_part+
substr_part: _value | "*"

attr: /[a-zA-Z][a-zA-Z0-9-;]*/ | oid
?oid: /\d+[\.\d]*/
!dn: "dn"i
matchingrule: attr

_value: /([^\x00()*\\]|\\\\[0-9a-fA-F]{2}|\\[0-9a-fA-F]{2})+/
value: ows _value? ows

ows: ws?
ws: WS
%import common.WS
%import common.HEXDIGIT
"""


class EscapeMode(enum.IntEnum):
    """Escape mode."""

    SPECIAL = 0
    r"""Escapes only the special characters defined in RFC 4515 (\, *, (, ), \x00)."""

    RESTRICTED = 1
    """
    Escapes all characters except for safe ASCII alphanumerics (0-9, A-Z, a-z)
    and a limited set of punctuation. Intended to allow only a small, controlled character set in filter values.
    """

    ALL = 2
    """Escapes every character, including all ASCII characters. Produces a fully hexadecimal-escaped filter value."""


class WalkStrategy(enum.IntEnum):
    """Walk stategy."""

    BOTH = 0
    POST = 1
    PRE = 2


class Token(str):  # noqa: FURB189
    __slots__ = ()


class _Value(str):  # noqa: FURB189
    __slots__ = ('prefix', 'suffix')


class Attribute:
    """An LDAP attribute."""

    __slots__ = ('attribute', 'dn', 'eq', 'matchingrule')

    ALLOWED_CHARS = string.ascii_letters + string.digits + ';-.'
    ALLOWED_OID = string.digits + '.'

    @property
    def approx(self) -> Self:
        return type(self)(self.attribute, self.dn, self.matchingrule, _equality_method=Filter.get_approx)

    def extensible(self, dn: str | None = None, matchingrule: str | None = None) -> Self:
        return type(self)(self.attribute, dn, matchingrule)

    def __init__(self, attribute: str, dn: str | None = None, matchingrule: str | None = None, *, _equality_method=None):
        self.attribute = attribute
        self.dn = dn
        self.matchingrule = matchingrule
        if _equality_method:
            self.eq = _equality_method
        elif dn or matchingrule:
            self.eq = self.eq_ext
        else:
            self.eq = Filter.get_eq
        if attribute.strip(self.ALLOWED_CHARS) or (attribute[0].isdigit() and attribute.strip(self.ALLOWED_OID)):
            raise ValueError(attribute)
        if dn and dn.lower() != 'dn':
            raise ValueError(dn)
        if matchingrule and (matchingrule.strip(self.ALLOWED_CHARS) or (matchingrule[0].isdigit() and matchingrule.strip(self.ALLOWED_OID))):
            raise ValueError(matchingrule)

    def eq_ext(self, attr, value):
        return Filter.get_extensible(attr, self.dn, self.matchingrule, value)

    def __eq__(self, other: Sequence):
        if isinstance(other, (list, tuple)):
            return Filter.get_substring(self.attribute, *other)
        return self.eq(self.attribute, other)

    def __ne__(self, other: Sequence | None):
        if other is None:
            return Filter.get_pres(self.attribute)
        return Filter.get_not(self == other)

    def __gt__(self, other: str | int):
        return Filter.get_gt(self.attribute, other)

    def __ge__(self, other: str | int):
        return Filter.get_gt_eq(self.attribute, other)

    def __lt__(self, other: str | int):
        return Filter.get_lt(self.attribute, other)

    def __le__(self, other: str | int):
        return Filter.get_lt_eq(self.attribute, other)

    def __hash__(self) -> str:
        return hash((self.attribute, self.dn, self.matchingrule))


class Expression:
    """Base class for any expression."""

    __slots__ = ()

    def __or__(self, other: Sequence):
        expr = other
        if not isinstance(other, Sequence):
            expr = [expr]
        return Filter.get_or(self, *expr)

    def __and__(self, other: Sequence):
        expr = other
        if not isinstance(other, Sequence):
            expr = [expr]

        return Filter.get_and(self, *expr)

    def negate(self):
        return Filter.get_not(self)


class Comparison(Expression):
    """Base class for comparison operators."""

    __slots__ = ('_end', '_extra', '_lead', '_mid', '_trail', 'attr', 'is_escaped', 'raw_value')
    operator = None
    expression = ''

    @property
    def value(self) -> str:
        """Get the value in a decoded form."""
        if self.is_escaped:
            return Filter.unescape(self.raw_value)
        return self.raw_value

    @value.setter
    def value(self, raw_value: str) -> None:
        self.raw_value = raw_value

    @property
    def escaped(self) -> str:
        """Get the value in a encoded/escaped form."""
        if self.is_escaped:
            return self.raw_value
        value = self.raw_value
        if value != value.strip():
            return Filter.escape(value, EscapeMode.RESTRICTED)
        return Filter.escape(value, EscapeMode.SPECIAL)

    def __init__(self, attr: str, value: str, is_escaped: bool = False):
        self.attr = attr
        self.raw_value = value
        self.is_escaped = is_escaped
        self._extra = ''
        self._lead = ''
        self._mid = ''
        self._trail = ''
        self._end = ''

    def copy(self) -> Self:
        """Copy the object (without preserving optional whitespace)."""
        return type(self)(self.attr, self.raw_value, is_escaped=self.is_escaped)

    def __str__(self):
        return f'{self._lead}{self.attr}{self._extra}{self.expression}{self._mid}{self.escaped}{self._end}{self._trail}'

    def __repr__(self):
        val = repr(self.raw_value) if self.is_escaped else f'escape({self.raw_value!r})'
        if isinstance(self, PresenceMatch):
            val = ''
        return f'{type(self).__name__}({self.attr}{self._extra}{self.expression}{val})'

    def __hash__(self):
        return hash((self.expression, self.attr, self.value))

    def __eq__(self, other):
        if isinstance(other, Comparison):
            return (self.expression, self.attr, self.value) == (other.expression, other.attr, other.value)
        return NotImplemented


class EqualityMatch(Comparison):
    """Compare for equality (attr=value)."""

    operator = operator.eq
    expression = '='


class GreaterOrEqual(Comparison):
    """Compare for greater or equals (attr>=value)."""

    operator = operator.ge
    expression = '>='


class LessOrEqual(Comparison):
    """Compare for less than or equals (attr<=value)."""

    operator = operator.le
    expression = '<='


class ApproximateMatch(Comparison):
    """Compare approximately (attr~=value)."""

    expression = '~='


class SubstringMatch(Comparison):
    """Compare substring match (attr=val*)."""

    operator = operator.contains
    expression = '='

    @property
    def values(self):
        """Get substring match values."""
        return tuple(self.value.split('*'))


class PresenceMatch(Comparison):
    """Compare for presence (attr=*)."""

    expression = '=*'


class ExtensibleMatch(Comparison):
    """Compare with extensible match (attr:dn:rule:=value, optional: dn / Matching Rule OID)."""

    expression = ':='
    __slots__ = (*Comparison.__slots__, 'dn', 'matchingrule')

    def copy(self) -> Self:
        """Copy the object (without preserving optional whitespace)."""
        return type(self)(self.attr, self.value, self.dn, self.matchingrule, is_escaped=self.is_escaped)

    def __init__(self, attr, value, dn, matchingrule, is_escaped=True):
        super().__init__(attr, value, is_escaped=is_escaped)
        extra = ''
        if dn:
            extra = f':{dn}'
        if matchingrule:
            extra = f'{extra}:{matchingrule}'
        self._extra = extra
        self.dn = dn
        self.matchingrule = matchingrule


class Container(Expression):
    """A group of one expression without brackets."""

    __slots__ = ('_expressions', '_sep')
    operator = None
    expression = ''

    @property
    def operators(self) -> tuple[Self]:
        """Get all operator expressions."""
        return tuple(op for op in self._expressions if isinstance(op, (AND, OR, NOT)))

    @property
    def comparisons(self) -> tuple[Comparison]:
        """Get all comparison expressions."""
        return tuple(op for op in self._expressions if isinstance(op, Comparison))

    @property
    def expressions(self) -> tuple[Comparison | Self]:
        """Get all comparison or operator expressions."""
        return tuple(op for op in self._expressions if isinstance(op, (Comparison, AND, OR, NOT)))

    def __init__(self, /, expressions: Sequence[Expression], *, sep: Token | None = None):
        self._expressions = expressions
        self._sep = sep

    def copy(self) -> Self:
        """Copy the object (without preserving optional whitespace)."""
        return type(self)([e.copy() for e in self._expressions])

    def append(self, expression: Expression) -> None:
        """Append to the operator list."""
        self._expressions.append(expression)

    def insert(self, expression: Expression, index: int = 0) -> None:
        """Insert into the operator list."""
        self._expressions.insert(index, expression)

    def replace(self, expression: Expression, replacement: Expression) -> None:
        """Insert into the operator list."""
        index = next((i for i, e in enumerate(self._expressions) if e is expression), None)
        if index is None:  # pragma: no cover
            raise ValueError('Not in expressions', expression)  # noqa: TRY003
        self._expressions[index] = replacement

    def remove(self, expression: Expression) -> None:
        """Insert from the operator list."""
        index = next((i for i, e in enumerate(self._expressions) if e is expression), None)
        if index is None:  # pragma: no cover
            raise ValueError('Not in expressions', expression)  # noqa: TRY003
        self._expressions.pop(index)

    def __str__(self):
        return ''.join(str(expr) for expr in self._expressions)

    def __repr__(self):
        exprs = f' {self.expression} '.join(repr(x) for x in self._expressions)
        return f'{type(self).__name__}( {exprs} )'


class Group(Container):
    """A group of one expression within brackets."""

    def __str__(self):
        sep = self._sep or ''
        groups = ''.join(f'({expr})' if isinstance(expr, Comparison) else str(expr) for expr in self._expressions)
        return f'({self.expression}{sep}{groups})' if self.expression else groups


class Operator(Group):
    """A logical operator."""


class AND(Operator):
    """A group of AND conjunction expressions ``(&(...)(...))``."""

    operator = operator.and_
    expression = '&'


class OR(Operator):
    """A group of OR disjunction expressions ``(|(...)(...))``."""

    operator = operator.or_
    expression = '|'


class NOT(Operator):
    """A group of NOT negation expressions ``(!( ... ))``."""

    operator = operator.not_
    expression = '!'


class Filter:
    """A LDAP Filter according to RFC 4515."""

    __slots__ = ('_debug', '_tree', 'ast', 'filter_expr')

    parser = Lark(LDAP_FILTER_GRAMMAR)
    RE_HEXESCAPE = re.compile(r'\\([0-9A-Fa-f]{2})')

    def __init__(self, /, filter_expr: str | None, *, strict=False, _debug=False) -> None:
        # TODO: security: restrict length
        # TODO: security: restrict depth
        # TODO: security: restrict number of escape sequences
        self.filter_expr = filter_expr
        self.ast: Container
        self._tree = None
        self._debug = _debug
        self.parse(strict)

    @property
    def root(self) -> Expression | None:
        """The first object in the filter."""
        if self._tree is None:  # pragma: no cover
            return None
        for expr in self.ast._expressions:
            if isinstance(expr, Token):
                continue  # pragma: no cover
            return expr
        return expr  # pragma: no cover

    def parse(self, strict: bool = False) -> None:
        """Parse."""
        if not self.filter_expr or self.filter_expr == ' ':
            self.ast = Container([])
            return
        transformer = _FilterTransformer(strict)
        try:
            self._tree = self.parser.parse(self.filter_expr)
            root = transformer.transform(self._tree)
            self.ast = Container([root])
        except lark.exceptions.LarkError:
            if self._debug:  # pragma: no cover
                raise
            raise self.error() from None
        if strict:
            self.walk(self._disallow_whitespace, self._disallow_whitespace)

    def _disallow_whitespace(self, fil: Self, parent: Expression, expr: Expression):  # noqa: ARG002
        if isinstance(expr, Comparison):
            if expr._mid or expr._lead or expr._trail or expr._end:
                raise self.error()
        elif isinstance(expr, Operator) and expr._sep:  # pragma: no cover
            raise self.error()  # currently impossible, broken OWS

    def error(self) -> errors.FilterError:
        """Get FilterError."""
        return errors.FilterError({'result': -7, 'desc': 'Bad search filter', 'info': str(self.filter_expr), 'ctrls': []})

    def pretty(self, indent: int = 0) -> str:
        """Transform into a pretty presentation."""

        def _pretty(expr, level):
            pad = '  ' * level
            if isinstance(expr, Comparison):
                return f'{pad}({expr})'
            if isinstance(expr, (AND, OR)):
                head = f'{pad}({expr.expression}'
                children = '\n'.join(_pretty(e, level + 1) for e in expr._expressions if not isinstance(e, Token))
                return f'{head}\n{children}\n{pad})'
            if isinstance(expr, NOT):
                head = f'{pad}({expr.expression}'
                children = ''.join(_pretty(e, 0) for e in expr._expressions if not isinstance(e, Token))
                return f'{head}{children})'
            if isinstance(expr, (Group, Container)):
                return '\n'.join(_pretty(e, level) for e in expr._expressions if not isinstance(e, Token))
            raise TypeError(f'Unexpected expression: {type(expr)}')  # noqa: TRY003,EM102  # pragma: no cover

        if self.root is None:
            return ''
        return _pretty(self.root, indent)

    @classmethod
    def attr(cls, attr: str, dn: str | None = None, matchingrule: str | None = None) -> Attribute:
        """Get attribute for comparison."""
        return Attribute(attr, dn, matchingrule)

    @classmethod
    def get_eq(cls, attr: str, value: str) -> EqualityMatch:
        """Get EqualityMatch."""
        return EqualityMatch(attr, value, is_escaped=False)

    @classmethod
    def get_approx(cls, attr: str, value: str) -> ApproximateMatch:
        """Get ApproximateMatch."""
        return ApproximateMatch(attr, value, is_escaped=False)

    @classmethod
    def get_pres(cls, attr: str) -> PresenceMatch:
        """Get PresenceMatch."""
        return PresenceMatch(attr, '', is_escaped=False)

    @classmethod
    def get_substring(cls, attr: str, *values: str) -> SubstringMatch:
        """Get SubstringMatch."""
        return SubstringMatch(attr, ''.join(values), is_escaped=False)

    @classmethod
    def get_gt_eq(cls, attr: str, value: str | int) -> GreaterOrEqual:
        """Get GreaterOrEqual."""
        return GreaterOrEqual(attr, str(value), is_escaped=False)

    @classmethod
    def get_gt(cls, attr: str, value: str | int) -> NOT | GreaterOrEqual:
        """Get greather then equivialent."""
        if isinstance(value, str):
            return cls.get_not(cls.get_lt_eq(attr, value))
        return cls.get_gt_eq(attr, value + 1)

    @classmethod
    def get_lt_eq(cls, attr: str, value: str | int) -> LessOrEqual:
        """Get LessOrEqual."""
        return LessOrEqual(attr, str(value), is_escaped=False)

    @classmethod
    def get_lt(cls, attr: str, value: str | int) -> NOT | LessOrEqual:
        """Get Lower than equivialent."""
        if isinstance(value, str):
            return cls.get_not(cls.get_gt_eq(attr, value))
        return cls.get_lt_eq(attr, value - 1)

    @classmethod
    def get_extensible(cls, attr: str, dn: str | None, matchingrule: str | None, value: str) -> ExtensibleMatch:
        """Get ExtensibleMatch."""
        return ExtensibleMatch(attr, value, dn, matchingrule, is_escaped=False)

    @classmethod
    def get_not(cls, expression: Expression) -> NOT:
        """Get negation."""
        return NOT([expression])

    @classmethod
    def get_and(cls, *expressions: Expression) -> AND:
        """Get conjunction."""
        return AND(expressions)

    @classmethod
    def get_or(cls, *expressions: Expression) -> OR:
        """Get disjunction."""
        return OR(expressions)

    @classmethod
    def from_format(cls, format_string, values) -> Self:
        """Get a Filter from a filter format string."""
        return cls(cls.escape_formatted(format_string, values))

    @classmethod
    def escape(cls, value: str, escape_mode: EscapeMode = EscapeMode.RESTRICTED) -> str:
        """Escape LDAP filter characters."""
        return ldap.filter.escape_filter_chars(value, escape_mode)

    @classmethod
    def unescape(cls, value: str) -> str:
        """Reverse the escaping of filter characters."""
        return cls.RE_HEXESCAPE.sub(
            lambda m: bytes([int(m.group(1), 16)]).decode('utf-8', errors='replace'),
            value,
        )

    @classmethod
    def escape_formatted(cls, format_string: str, values: Sequence[str]) -> str:
        """Escape LDAP filter characters in format string."""
        return ldap.filter.filter_format(format_string, values)

    @classmethod
    def time_span_filter(cls, from_timestamp=0, until_timestamp=None, delta_attr='modifyTimestamp') -> Self:
        """Get timespan filter e.g. '(&(modifyTimestamp>=19700101000000Z)(!(modifyTimestamp>=19700101000001Z)))'."""
        return cls(ldap.filter.time_span_filter('', from_timestamp, until_timestamp, delta_attr))

    def walk(self, comparison_callback: Callable | None, operator_callback: Callable | None, strategy: WalkStrategy = WalkStrategy.POST) -> None:
        """Walk the filter expressions and operator conjunctions iteratively."""
        stack = deque([(self.ast, self.root, True)])

        while stack:
            parent, expression, is_pre_visit = stack.pop()

            if isinstance(expression, Comparison):
                if comparison_callback:
                    comparison_callback(self, parent, expression)
            elif isinstance(expression, (Operator, Group, Container)):
                if is_pre_visit:
                    if operator_callback and strategy in {WalkStrategy.PRE, WalkStrategy.BOTH} and isinstance(expression, (AND, OR, NOT)):
                        operator_callback(self, parent, expression)

                    if operator_callback and strategy in {WalkStrategy.POST, WalkStrategy.BOTH} and isinstance(expression, (AND, OR, NOT)):
                        stack.append((parent, expression, False))

                    for child_expression in expression.expressions[::-1]:
                        stack.append((expression, child_expression, True))
                else:
                    operator_callback(self, parent, expression)
            elif isinstance(expression, Token):  # pragma: no cover
                pass  # not possible in the root!?
            else:  # pragma: no cover
                raise TypeError(expression)

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.root!r})'

    def __str__(self) -> str:
        if self._tree is None:
            return self.filter_expr
        return str(self.ast)

    # def _pretty(self):
    #     return self._tree.pretty()


@v_args(inline=True)
class _FilterTransformer(Transformer):
    """Filter tree Transformer."""

    def __init__(self, strict):
        self.__strict = strict
        super().__init__()

    def start(self, value):  # noqa: PLR6301
        if isinstance(value, Comparison):
            return Group([value])
        if isinstance(value, Container):
            return value
        raise TypeError(type(value))  # pragma: no cover

    def bare(self, cmp):  # noqa: PLR6301
        return Container([cmp])

    def groups(self, *groups):  # noqa: PLR6301
        return list(groups)

    def expression(self, ld, cmp, tr):  # noqa: PLR6301
        cmp._lead = ld or ''
        cmp._trail = tr or ''
        return cmp

    def and_(self, sep, ld, exprs, tr):
        return AND(self._filter([ld, *list(exprs), tr]), sep=sep)

    def or_(self, sep, ld, exprs, tr):
        return OR(self._filter([ld, *list(exprs), tr]), sep=sep)

    def not_(self, sep, ld, expr, tr):
        return NOT(self._filter([ld, expr, tr]), sep=sep)

    def _filter(self, items):  # noqa: PLR6301
        return [item for item in items if item is not None]

    def attr(self, attr):  # noqa: PLR6301
        return str(attr)

    def value(self, ld, *value) -> _Value:  # noqa: PLR6301
        if len(value) > 1:
            val, tr = value
        else:
            val, tr = '', value[0]

        if val != val.strip():
            ld_, val, tr_ = val.partition(val.strip())
            ld = ld or ld_
            tr = tr or tr_
        val = _Value(val)
        val.prefix = ld
        val.suffix = tr
        return val

    def _prefix(self, data: _Value, value):  # noqa: PLR6301
        value._mid = data.prefix or ''
        value._end = data.suffix or ''
        return value

    def equality(self, attr, value):
        return self._prefix(value, EqualityMatch(str(attr), str(value), is_escaped=True))

    def presence(self, attr, value=''):
        return self._prefix(self.value(None, value, None), PresenceMatch(str(attr), '', is_escaped=True))

    def substring(self, attr, value):
        value = self.value(None, value, None)
        if value == '*':
            return self.presence(attr, value)
        return self._prefix(value, SubstringMatch(attr, value, is_escaped=True))

    def substr_part(self, value='*'):  # noqa: PLR6301
        return value

    def substrings(self, *values):  # noqa: PLR6301
        value = ''.join(values)
        if '**' in value:
            raise ValueError()
        return value

    def ge(self, attr, value):
        return self._prefix(value, GreaterOrEqual(attr, str(value), is_escaped=True))

    def le(self, attr, value):
        return self._prefix(value, LessOrEqual(attr, str(value), is_escaped=True))

    def extmatch_attr_nodn_match(self, attr, matchingrule, value):
        return self._prefix(value, ExtensibleMatch(attr, str(value), '', matchingrule, is_escaped=True))

    def extmatch_attr_dn_nomatch(self, attr, dn, value):
        return self._prefix(value, ExtensibleMatch(attr, str(value), dn, '', is_escaped=True))

    def extmatch_noattr_nodn_match(self, matchingrule, value):
        if matchingrule.lower() == 'dn':
            raise ValueError()
        return self._prefix(value, ExtensibleMatch('', str(value), '', matchingrule, is_escaped=True))

    def extmatch_attr_nodn_nomatch(self, attr, value):
        return self._prefix(value, ExtensibleMatch(attr, str(value), '', '', is_escaped=True))

    def extmatch_attr_dn_match(self, attr, dn, matchingrule, value):
        return self._prefix(value, ExtensibleMatch(attr, str(value), dn, matchingrule, is_escaped=True))

    def extmatch_noattr_dn_match(self, dn, matchingrule, value):
        return self._prefix(value, ExtensibleMatch('', str(value), dn, matchingrule, is_escaped=True))

    def approx(self, attr, value):
        return self._prefix(value, ApproximateMatch(str(attr), str(value), is_escaped=True))

    def dn(self, value='dn'):  # noqa: PLR6301
        return str(value)

    def matchingrule(self, value):  # noqa: PLR6301
        return str(value)

    def ws(self, value):  # noqa: PLR6301
        return Token(value)

    def ows(self, value=None):  # noqa: PLR6301
        if value is None:
            return None
        return Token(value)

    def __default__(self, data, children, meta):  # noqa: PLW3201
        if data == 'and':
            return self.and_(*children)
        if data == 'or':
            return self.or_(*children)
        if data == 'not':
            return self.not_(*children)
        return super().__default__(data, children, meta)  # pragma: no cover



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""LDAP Attributes."""

import contextlib
import typing


class Attributes(dict):  # noqa: FURB189
    """LDAP Attributes."""

    ALIASES: typing.ClassVar = {}
    SCHEMA = None

    def __getitem__(self, key):
        with contextlib.suppress(KeyError):
            return super().__getitem__(key)
        return self.__missing__(key)

    def __missing__(self, key):
        key = self.ALIASES.get(key, key)
        return {k.lower(): v for k, v in self.items()}[key.lower()]

    @classmethod
    def set_schema(cls, subschema):
        """Set aliases from schema."""
        cls.SCHEMA = subschema
        cls.ALIASES.update(subschema.get_attribute_aliases())



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""Data wrapper."""

from dataclasses import dataclass
from typing import Any, TypeAlias

import ldap.controls

from freeiam.ldap.attr import Attributes
from freeiam.ldap.constants import ResponseType
from freeiam.ldap.dn import DN


LDAPControlList: TypeAlias = list[ldap.controls.LDAPControl]


@dataclass
class Controls:
    """The LDAP request controls."""

    server: LDAPControlList | None = None
    client: LDAPControlList | None = None
    response: LDAPControlList | None = None

    def get(self, control):
        """Get the control from the list of response controls."""
        for ctrl in self.response or []:
            if ctrl.controlType == control.controlType:
                return ctrl
        return None

    @classmethod
    def expand(cls, controls):
        if controls is None:
            return {}
        return {'serverctrls': controls.server, 'clientctrls': controls.client}

    @classmethod
    def append_server(cls, controls, control):
        controls = Controls([]) if controls is None else controls
        controls.server.append(control)
        return controls

    @classmethod
    def set_server(cls, controls, control):
        controls = Controls([]) if controls is None else controls
        controls.server = [ctrl for ctrl in controls.server if ctrl.controlType != control.controlType] + [control]
        return controls


@dataclass
class _Response:
    """The raw response of ldapobject.result4()."""

    type: ResponseType | None
    data: list | None
    msgid: int | None
    ctrls: list[Any] | None
    name: Any
    value: Any

    def __post_init__(self):
        if not isinstance(self.type, ResponseType | None):
            self.type = ResponseType(self.type)


@dataclass
class Page:
    """A page of a paginated search result."""

    page: int
    """The current page number (starting at one)."""

    entry: int
    """The number of the current entry on this page (starting at one)."""

    page_size: int
    """The number of entries per page."""

    results: int | None = None
    """The total number of search results."""

    last_page: int | None = None
    """The last page of all search results."""

    @property
    def is_last_in_page(self) -> bool:
        """Whether this is the last entry on the current page."""
        return self.page_size == self.entry


@dataclass
class Result:
    """The wrapped result of an operation. Allows accessing response controls."""

    dn: DN
    """The new or unchanged DN of the object."""

    attr: Attributes
    """The result LDAP attributes, if the operation provides some."""

    controls: Controls | None
    """LDAP response controls."""

    _response: _Response
    """The raw LDAP result."""

    page: Page | None = None
    """The page of a paginated search result."""

    @classmethod
    def from_response(cls, dn, attr, controls, response, **kwargs):
        dn = dn if dn is None else DN.get(dn)
        attr = attr if attr is None else Attributes(attr)
        return cls(dn, attr, cls._control_response(controls, response.ctrls), response, **kwargs)

    @classmethod
    def set_controls(cls, response: _Response, controls: Controls | None):
        if controls is None:
            return
        controls.response = response.ctrls

    @classmethod
    def _control_response(cls, controls, response_ctrls):
        if controls is None:
            return Controls(None, None, response_ctrls)

        return Controls(controls.server and controls.server.copy(), controls.client and controls.client.copy(), response_ctrls)



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""LDAP Schemata."""

from ldap.schema import AttributeType, ObjectClass


class Schema:
    """LDAP Schemata."""

    __slots__ = ('_schema',)

    def __init__(self, schema):
        self._schema = schema

    def get_object_class(self, name):
        """Get object class by name."""
        for oc in self.get_object_classes():
            if name in oc.names:
                return oc
        return None

    def get_object_class_by_oid(self, oid):
        """Get object class by OID."""
        return self._schema.get_obj(ObjectClass, oid)

    def get_object_classes(self):
        """Get all object classes."""
        for oid in self._schema.listall(ObjectClass):
            yield self.get_object_class_by_oid(oid)

    def get_attribute(self, name):
        """Get attribute by name."""
        for attr in self.get_attributes():
            if name in attr.names:
                return attr
        return None

    def get_attributes(self):
        """Get all attributes."""
        for oid in self._schema.listall(AttributeType):
            yield self._schema.get_obj(AttributeType, oid)

    def get_attribute_aliases(self) -> dict:
        """Get aliases of attribute names."""
        return {
            alias: attr.names[0]
            for attr in self.get_attributes()
            for alias in attr.names[1:]
            if len(attr.names) > 1
        }  # fmt: skip



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""LDAP Constants."""

# See https://app.readthedocs.org/projects/python-ldap/downloads/pdf/latest/

from enum import IntEnum
from typing import TypeAlias

import ldap


def _getoptional(name, default):
    return getattr(ldap, name, default)


class Scope(IntEnum):
    """All possible search scopes."""

    Base = ldap.SCOPE_BASE
    """Base entry scope"""

    Onelevel = ldap.SCOPE_ONELEVEL
    """Direct children scope"""

    Subtree = ldap.SCOPE_SUBTREE
    """Whole subtree scope"""

    Subordinate = _getoptional('SCOPE_SUBORDINATE', 3)

    One = Onelevel
    Sub = Subtree

    BASE = Base
    ONELEVEL = Onelevel
    SUBTREE = Subtree


class Mod(IntEnum):
    """Modification list entry."""

    Add = ldap.MOD_ADD
    BinaryValues = ldap.MOD_BVALUES
    Delete = ldap.MOD_DELETE
    Increment = ldap.MOD_INCREMENT
    Replace = ldap.MOD_REPLACE


class Version(IntEnum):
    """LDAP Protocol Version."""

    LDAPV1 = ldap.VERSION1
    """Version 1"""

    LDAPV2 = ldap.VERSION2
    """Version 2"""

    LDAPV3 = ldap.VERSION3
    """Version 3"""

    Max = ldap.VERSION_MAX
    """Maximum protocol version"""

    Min = ldap.VERSION_MIN
    """Minimum protocol version"""

    V1 = LDAPV1
    V2 = LDAPV2
    V3 = LDAPV3


class Option(IntEnum):
    """LDAP Options."""

    # API / Info
    ApiFeatureInfo = ldap.OPT_API_FEATURE_INFO
    """Returns API feature information (`int`)."""

    ApiInfo = ldap.OPT_API_INFO
    """Returns API information (`int`)."""

    # Debug / Diagnostics
    DebugLevel = ldap.OPT_DEBUG_LEVEL
    """Sets the debug level within the underlying OpenLDAP C library (`int`).

    libldap sends the log messages to stderr.
    """

    DiagnosticMessage = ldap.OPT_DIAGNOSTIC_MESSAGE
    """Gets the diagnostic message of the last operation (`str`)."""

    ErrorNumber = ldap.OPT_ERROR_NUMBER
    """Gets the error number of the last occurred error (`int`)."""

    ErrorString = ldap.OPT_ERROR_STRING
    """Gets the error string corresponding to the last occurred error (`str`)."""

    # Connection / Network
    FileDescriptor = ldap.OPT_DESC
    """Gets the file descriptor of the LDAP connection (`int`)."""

    ConnectAsync = _getoptional('OPT_CONNECT_ASYNC', 20496)
    """Enables asynchronous connect mode (`int`)."""

    HostName = ldap.OPT_HOST_NAME
    """The hostname used for the connection (`str`)."""

    NetworkTimeout = ldap.OPT_NETWORK_TIMEOUT
    """Network timeout in seconds (`int`).

    A timeout of -1 or None resets the timeout to infinity.
    """

    TCPUserTimeout = _getoptional('OPT_TCP_USER_TIMEOUT', 20501)
    """TCP user timeout in milliseconds (`int`)."""

    Timeout = ldap.OPT_TIMEOUT
    """Operation timeout in seconds (`int`).

    A timeout of -1 or None resets the timeout to infinity.
    """

    # LDAP behavior
    ProtocolVersion = ldap.OPT_PROTOCOL_VERSION
    """LDAP protocol version to use for the connection (`Version`)."""

    Dereference = ldap.OPT_DEREF
    """Specifies how alias dereferencing is performed (`Dereference`)."""

    Referrals = _getoptional('OPT_REFERRALS', 8)
    """Specifies whether referrals should be automatically chased (`int`)."""

    Refhoplimit = ldap.OPT_REFHOPLIMIT
    """Maximum number of referral hops (`int`)."""

    Restart = ldap.OPT_RESTART
    """Specifies whether operations are automatically restarted (`int`)."""

    DefaultBase = _getoptional('OPT_DEFBASE', 20489)
    """Default search base DN for operations (`str`)."""

    URI = ldap.OPT_URI
    """LDAP URI(s) for the connection (`str`)."""

    # Result / Limits
    ResultCode = ldap.OPT_RESULT_CODE
    """Gets the result code of the last operation (`int`)."""

    MatchedDN = ldap.OPT_MATCHED_DN
    """Gets the matched distinguished name from the last operation (`str`)."""

    Sizelimit = ldap.OPT_SIZELIMIT
    """Specifies the maximum number of entries to return for a search (`int`)."""

    Timelimit = ldap.OPT_TIMELIMIT
    """Specifies the maximum time in seconds a search may run (`int`)."""

    # Controls
    ClientControls = ldap.OPT_CLIENT_CONTROLS
    """List of LDAP client controls (`list`)."""

    ServerControls = ldap.OPT_SERVER_CONTROLS
    """List of LDAP server controls (`list`)."""


class SASLOption(IntEnum):
    """SASL Options (must be set per connection)."""

    AuthCID = ldap.OPT_X_SASL_AUTHCID
    AuthZID = ldap.OPT_X_SASL_AUTHZID
    Mechanism = ldap.OPT_X_SASL_MECH
    NoCanonicalization = _getoptional('OPT_X_SASL_NOCANON', 24843)
    """If set to zero, SASL host name canonicalization is disabled."""
    Realm = ldap.OPT_X_SASL_REALM
    Secprops = ldap.OPT_X_SASL_SECPROPS
    SSF = ldap.OPT_X_SASL_SSF
    """Security Strength Factor"""
    SSFExternal = ldap.OPT_X_SASL_SSF_EXTERNAL
    SSFMax = ldap.OPT_X_SASL_SSF_MAX
    """Maximum Security Strength Factor"""
    SSFMin = ldap.OPT_X_SASL_SSF_MIN
    """Minimum Security Strength Factor"""
    Username = _getoptional('OPT_X_SASL_USERNAME', 24844)
    """SASL Username"""


class OptionValue(IntEnum):
    """LDAP Option Values."""

    Off = ldap.OPT_OFF
    On = ldap.OPT_ON
    Success = ldap.OPT_SUCCESS

    NoLimit = ldap.NO_LIMIT


class KeepAlive(IntEnum):
    """Keep Alive Option values."""

    Idle = _getoptional('OPT_X_KEEPALIVE_IDLE', 25344)
    Interval = _getoptional('OPT_X_KEEPALIVE_INTERVAL', 25346)
    Probes = _getoptional('OPT_X_KEEPALIVE_PROBES', 25345)


class TLSOption(IntEnum):
    """TLS Options."""

    # Certificate files and directories
    CACertdir = ldap.OPT_X_TLS_CACERTDIR
    """Path to a directory containing CA certificates (`str`)."""

    CACertfile = ldap.OPT_X_TLS_CACERTFILE
    """Path to a CA certificate file (`str`)."""

    Certfile = ldap.OPT_X_TLS_CERTFILE
    """Path to the client/server certificate file (`str`)."""

    Keyfile = ldap.OPT_X_TLS_KEYFILE
    """Path to the private key file corresponding to the certificate (`str`)."""

    DHFile = ldap.OPT_X_TLS_DHFILE
    """Path to the Diffie-Hellman parameters file (`str`)."""

    ECName = _getoptional('OPT_X_TLS_ECNAME', 24594)
    """Name of the elliptic curve to use (`str`)."""

    # Certificate verification
    CRLCheck = _getoptional('OPT_X_TLS_CRLCHECK', 24587)
    """Certificate Revocation List check policy (`TLSCRLCheck`)."""

    CRLFile = _getoptional('OPT_X_TLS_CRLFILE', 24592)
    """Path to a CRL file (`str`)."""

    RequireCert = ldap.OPT_X_TLS_REQUIRE_CERT
    """Certificate requirement level (`TLSRequireCert`)."""

    RequireSAN = _getoptional('OPT_X_TLS_REQUIRE_SAN', 24602)
    """Requirement for Subject Alternative Name (`TLSRequireCert`)."""

    PeerCert = _getoptional('OPT_X_TLS_PEERCERT', 24597)
    """Path to the peer certificate file (`str`)."""

    # Protocol and version settings
    ProtocolMin = _getoptional('OPT_X_TLS_PROTOCOL_MIN', 24583)
    """Minimum allowed TLS protocol (`TLSProtocol`)."""

    ProtocolMax = _getoptional('OPT_X_TLS_PROTOCOL_MAX', 24603)
    """Maximum allowed TLS protocol (`TLSProtocol`)."""

    Version = _getoptional('OPT_X_TLS_VERSION', 24595)
    """TLS library version (`str`)."""

    # Cipher settings
    Cipher = _getoptional('OPT_X_TLS_CIPHER', 24596)
    """Cipher specification (`str`)."""

    CipherSuite = ldap.OPT_X_TLS_CIPHER_SUITE
    """Cipher suite selection (`str`)."""

    # Misc / control
    Package = _getoptional('OPT_X_TLS_PACKAGE', 24593)
    """TLS package to use (`str`)."""

    NewContext = _getoptional('OPT_X_TLS_NEWCTX', 24591)
    """Create a new internal TLS context (`int`).

    libldap does not apply all TLS settings immediately.
    Use this option with value 0 to instruct libldap to apply
    pending TLS settings and create a new internal TLS context.
    """


class TLSCRLCheck(IntEnum):
    """Values for Certificate Revocation List checks (TLSOption.CRLCheck)."""

    None_ = ldap.OPT_X_TLS_CRL_NONE
    Peer = ldap.OPT_X_TLS_CRL_PEER
    All = ldap.OPT_X_TLS_CRL_ALL


class TLSRequireCert(IntEnum):
    """Values for Certificate requirement or Subject Alternative Name (TLSOption.RequireCert and TLSOption.RequireSAN)."""

    Never = ldap.OPT_X_TLS_NEVER
    Allow = ldap.OPT_X_TLS_ALLOW
    Try = ldap.OPT_X_TLS_TRY
    Demand = ldap.OPT_X_TLS_DEMAND
    Hard = ldap.OPT_X_TLS_HARD


class TLSProtocol(IntEnum):
    """Values for TLSOption.ProtocolMin / TLSOption.ProtocolMax."""

    SSL3 = _getoptional('OPT_X_TLS_PROTOCOL_SSL3', 0x300)
    TLS10 = _getoptional('OPT_X_TLS_PROTOCOL_TLS1_0', 0x301)
    TLS11 = _getoptional('OPT_X_TLS_PROTOCOL_TLS1_1', 0x302)
    TLS12 = _getoptional('OPT_X_TLS_PROTOCOL_TLS1_2', 0x303)
    TLS13 = _getoptional('OPT_X_TLS_PROTOCOL_TLS1_3', 0x304)


class Dereference(IntEnum):
    """Dereference options."""

    Always = ldap.DEREF_ALWAYS
    Never = ldap.DEREF_NEVER
    Searching = ldap.DEREF_SEARCHING
    Finding = ldap.DEREF_FINDING


AnyOption: TypeAlias = Option | SASLOption | TLSOption | int
AnyOptionValue: TypeAlias = OptionValue | TLSCRLCheck | TLSRequireCert | TLSProtocol | KeepAlive | Dereference | int | str


class DNFormat(IntEnum):
    """Used for DN-parsing functions."""

    LDAP = ldap.DN_FORMAT_LDAP
    LDAPV2 = ldap.DN_FORMAT_LDAPV2
    LDAPV3 = ldap.DN_FORMAT_LDAPV3
    DCE = ldap.DN_FORMAT_DCE
    UFN = ldap.DN_FORMAT_UFN
    ADCanonical = ldap.DN_FORMAT_AD_CANONICAL
    Mask = ldap.DN_FORMAT_MASK
    Pretty = ldap.DN_PRETTY
    Skip = ldap.DN_SKIP
    NoLeadTrailSpaces = ldap.DN_P_NOLEADTRAILSPACES
    NoSpaceAfterDN = ldap.DN_P_NOSPACEAFTERRDN
    Pedantic = ldap.DN_PEDANTIC


class AVA(IntEnum):
    """Attribute Value Assertion formats."""

    Binary = ldap.AVA_BINARY
    NonPrintable = ldap.AVA_NONPRINTABLE
    Null = ldap.AVA_NULL
    String = ldap.AVA_STRING


class ResponseType(IntEnum):
    """LDAP Response types."""

    Add = ldap.RES_ADD
    Any = ldap.RES_ANY
    Bind = ldap.RES_BIND
    Compare = ldap.RES_COMPARE
    Delete = ldap.RES_DELETE
    Extended = ldap.RES_EXTENDED
    Intermediate = ldap.RES_INTERMEDIATE
    Modify = ldap.RES_MODIFY
    ModRDN = ldap.RES_MODRDN
    SearchEntry = ldap.RES_SEARCH_ENTRY
    SearchReference = ldap.RES_SEARCH_REFERENCE
    SearchResult = ldap.RES_SEARCH_RESULT
    Unsolicited = ldap.RES_UNSOLICITED


class LDAPChangeType(IntEnum):
    """LDAP change type for PersistentSearchControl control."""

    Add = 1
    Delete = 2
    Modify = 4
    ModifyDN = 8



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""FreeIAM version."""

__all__ = ['__version__', '__version_tuple__', 'version', 'version_tuple']

from importlib.metadata import version as get_version


def _parse_version_tuple(ver: str) -> tuple[int, ...]:
    return tuple(int(part) for part in ver.split('.') if part.isdigit())


__version__ = version = get_version(__package__)
__version_tuple__ = version_tuple = _parse_version_tuple(__version__)



# SPDX-FileCopyrightText: 2025 Florian Best
# SPDX-License-Identifier: MIT OR Apache-2.0
"""LDAP Connection."""

import contextlib
import logging
import math
import time
from collections.abc import AsyncGenerator, Callable, Generator
from typing import Any, Self, TypeAlias

import ldap.ldapobject
import ldap.modlist
import ldap.sasl
from ldap.schema import SCHEMA_ATTRS

from freeiam import errors
from freeiam.ldap._wrapper import Page, Result, _Response
from freeiam.ldap.attr import Attributes
from freeiam.ldap.constants import (
    AnyOption,
    AnyOptionValue,
    Option,
    OptionValue,
    ResponseType,
    Scope,
    TLSCRLCheck,
    TLSOption,
    TLSProtocol,
    TLSRequireCert,
    Version,
)
from freeiam.ldap.controls import Controls, server_side_sorting, simple_paged_results, virtual_list_view
from freeiam.ldap.dn import DN
from freeiam.ldap.schema import Schema


__all__ = ('Connection',)

log = logging.getLogger(__name__)

LDAPObject: TypeAlias = ldap.ldapobject.SimpleLDAPObject
LDAPAddList: TypeAlias = list[tuple[str, list[bytes]]]
LDAPModList: TypeAlias = list[tuple[int, str, list[bytes] | None]]
Sorting: TypeAlias = list[str | tuple[str, str | None, bool]]


class Connection:
    """
    A LDAP Connection.

    :ivar str uri: The LDAP URI.
    :ivar int timelimit: The global timelimit.
    :ivar bool automatic_reconnect: Whether automatic reconnection is enabled.
    :ivar int max_connection_attempts: number of connection attempt on connection loss.
    :ivar float retry_delay: The retry delay (in seconds) between the reconnection attempts.
    """

    __slots__ = (
        '__conn_s',
        '__reconnects_counter',
        '__schema',
        '_conn',
        '_hide_parent_exception',
        '_last_auth_state',
        '_options',
        '_start_tls',
        'automatic_reconnect',
        'max_connection_attempts',
        'retry_delay',
        'timeout',
        'uri',
    )

    @property
    def conn(self) -> LDAPObject:
        """The underlying connection."""
        if self._conn is None:
            raise RuntimeError('not connected')  # noqa: TRY003
        return self._conn

    @property
    def fileno(self):
        """Get the file descriptor number of the active socket connection."""
        if not self._conn or not hasattr(self._conn, '_l'):
            return -1
        return self._conn.fileno()

    @property
    def connected(self) -> bool:
        """Whether the connection is established."""
        return self.fileno != -1

    def __init__(
        self,
        uri: str | None = '',
        *,
        start_tls: bool = False,
        timeout: int = -1,
        automatic_reconnect: bool = True,
        max_connection_attempts: int = 10,
        retry_delay: float = 0.0,
        _hide_parent_exception: bool = True,
        _conn: LDAPObject = None,
    ):
        self._conn = _conn
        self.uri = uri
        self.timeout = timeout
        self.automatic_reconnect = automatic_reconnect
        self.max_connection_attempts = max_connection_attempts
        self.retry_delay = retry_delay
        self._start_tls = start_tls
        self.__reconnects_counter = 0
        self.__schema = {}
        self._last_auth_state = None
        self._options = []
        self._hide_parent_exception = _hide_parent_exception
        self.__conn_s = None

    def __enter__(self) -> Self:
        """Initialize asynchronous connection."""
        self.connect()
        return self

    def __exit__(self, etype, exc, etraceback) -> None:
        """Close connection on shutdown."""
        self.unbind()
        self.disconnect()

    def get_option(self, option: AnyOption) -> int:
        """Get a LDAP connection option."""
        with errors.LdapError.wrap(self._hide_parent_exception):
            return self.conn.get_option(option)

    def set_option(self, option: AnyOption, value: AnyOptionValue, *, append=True) -> None:
        """Set a LDAP connection option."""
        if append:
            self._options.append((option, value))
        with errors.LdapError.wrap(self._hide_parent_exception):
            return self.conn.set_option(option, value)

    def set_controls(self, controls: Controls):
        """Set LDAP controls for all operations on this connection."""
        if controls.server is not None:
            self.set_option(Option.ServerControls, controls.server)
        if controls.client is not None:
            self.set_option(Option.ClientControls, controls.client)

    @property
    def protocol_version(self) -> Version:
        """Get the LDAP protocol version."""
        return Version(self.get_option(Option.ProtocolVersion))

    @protocol_version.setter
    def protocol_version(self, value: Version):
        """Set the LDAP protocol version."""
        return self.set_option(Option.ProtocolVersion, value)

    @property
    def timelimit(self) -> int:
        """Get the LDAP time limit."""
        return self.get_option(Option.Timelimit)

    @timelimit.setter
    def timelimit(self, value: int):
        """Set the LDAP time limit."""
        return self.set_option(Option.Timelimit, value)

    @property
    def network_timeout(self) -> int:
        """Get the LDAP network timeout."""
        return self.get_option(Option.NetworkTimeout)

    @network_timeout.setter
    def network_timeout(self, value: int):
        """Set the LDAP network timeout."""
        return self.set_option(Option.NetworkTimeout, value)

    @property
    def dereference(self) -> int:
        """Get the de-reference setting."""
        return self.get_option(Option.Dereference)

    @dereference.setter
    def dereference(self, value: int):
        """Set the de-reference setting."""
        return self.set_option(Option.Dereference, value)

    @property
    def follow_referrals(self) -> bool | None:
        """Follow referrals enabled."""
        follow = self.get_option(Option.Referrals)
        if follow == -1:
            return None
        return follow == OptionValue.On

    @follow_referrals.setter
    def follow_referrals(self, value: bool):
        """Enable following of referrals."""
        return self.set_option(Option.Referrals, OptionValue.On if value else OptionValue.Off)

    @property
    def sizelimit(self) -> int:
        """Get the sizelimit setting."""
        return self.get_option(Option.Sizelimit)

    @sizelimit.setter
    def sizelimit(self, value: int):
        """Set the sizelimit setting."""
        return self.set_option(Option.Sizelimit, value)

    @classmethod
    def set_tls(
        cls,
        *,
        ca_certfile: str | None = None,
        ca_certdir: str | None = None,
        certfile: str | None = None,
        keyfile: str | None = None,
        require_cert: TLSRequireCert = TLSRequireCert.Demand,
        require_san: TLSRequireCert | None = None,
        minimum_protocol: TLSProtocol | None = None,
        cipher_suite: str | None = None,
        crlfile: None = None,
        crl_check: TLSCRLCheck | None = None,
    ) -> None:
        """Set the TLS certificate settings globally."""
        for option, value in (
            (TLSOption.CACertfile, ca_certfile),
            (TLSOption.CACertdir, ca_certdir),
            (TLSOption.Certfile, certfile),
            (TLSOption.Keyfile, keyfile),
            (TLSOption.ProtocolMin, minimum_protocol),
            (TLSOption.CipherSuite, cipher_suite),
            (TLSOption.RequireCert, require_cert),
            (TLSOption.RequireSAN, require_san),
            (TLSOption.CRLFile, crlfile),
            (TLSOption.CRLCheck, crl_check),
        ):
            if value is not None:
                cls.set_global_option(option, value)

        # apply the pending TLS settings, create new context:
        cls.set_global_option(TLSOption.NewContext, 0)

    @classmethod
    def get_global_option(cls, option: AnyOption) -> int:
        """Get a LDAP connection option."""
        return ldap.get_option(option)

    @classmethod
    def set_global_option(cls, option: AnyOption, value: AnyOptionValue) -> None:
        """Set a global LDAP option."""
        ldap.set_option(option, value)

    def connect(self, fileno: bool | None = None) -> None:
        """Connect to the LDAP server."""
        if self.connected:
            raise RuntimeError('already connected')  # noqa: TRY003

        if self.automatic_reconnect:
            self._conn = ldap.ldapobject.ReconnectLDAPObject(
                self.uri,
                trace_level=0,
                trace_file=None,
                trace_stack_limit=None,
                fileno=fileno,
                retry_max=self.max_connection_attempts,
                retry_delay=self.retry_delay,
            )
        else:
            self._conn = ldap.ldapobject.SimpleLDAPObject(
                self.uri,
                trace_level=0,
                trace_file=None,
                trace_stack_limit=None,
                fileno=fileno,
            )
        self.__conn_s = None
        self.__reconnects_counter = 0
        if self._start_tls:
            self.start_tls()

    def disconnect(self) -> None:
        """Disconnect from LDAP server."""
        self._conn = None
        self.__conn_s = None

    def reconnect(self, *, force: bool = True) -> None:
        """Reconnect to the LDAP server."""
        with errors.LdapError.wrap(self._hide_parent_exception):
            self.conn.reconnect(self.uri, self.max_connection_attempts, self.retry_delay, force=force)
            # if self._start_tls:
            #     self.start_tls()
            self._restore_options()
            self._restore_auth_state()

    def start_tls(self) -> None:
        """Start TLS."""
        self._start_tls = True
        with errors.LdapError.wrap(self._hide_parent_exception):
            self.conn.start_tls_s()

    def get_schema(self, subschema_dn: DN | str | None = None) -> ldap.schema.subentry.SubSchema:
        """Get LDAP Schema."""
        conn = self.conn
        # cache schema by connection
        if isinstance(conn, ldap.ldapobject.ReconnectLDAPObject) and conn._reconnects_done > self.__reconnects_counter:
            self.__schema[subschema_dn] = None
            self.__reconnects_counter = self.conn._reconnects_done

        if not self.__schema.get(subschema_dn):
            try:
                subschemasubentry = self.get(subschema_dn, attrs=['subschemaSubentry']) if subschema_dn else self.get_root_dse(['subschemaSubentry'])
                try:
                    subschemasubentry_dn = subschemasubentry.attr['subschemaSubentry'][0].decode('UTF-8')
                except KeyError:  # pragma: no cover; impossible?
                    if subschema_dn:
                        return self.get_schema()
                    subschema = None
            except (errors.NoSuchObject, errors.NoSuchAttribute, errors.InsufficientAccess, errors.UndefinedType):
                subschema = None
            else:
                try:
                    subschema = (self.get(subschemasubentry_dn, SCHEMA_ATTRS, '(objectClass=subschema)')).attr
                except errors.NoSuchObject:  # pragma: no cover
                    subschema = None
            self.__schema[subschema_dn] = Schema(ldap.schema.SubSchema(subschema, 0))
            Attributes.set_schema(self.__schema[subschema_dn])
        return self.__schema[subschema_dn]

    def bind(self, authzid: str | None, password: str | None, *, controls: Controls | None = None) -> None:
        """Authenticate via plaintext credentials."""
        conn = self.conn
        self._last_auth_state = ('simple_bind_s', authzid, password)
        response = self._execute(conn, conn.simple_bind, authzid, password, **Controls.expand(controls))
        return Result.from_response(None, None, controls, response)

    def bind_external(self):  # pragma: no cover
        """Authenticate via EXTERNAL method e.g. UNIX socket or TLS client certificate."""
        with errors.LdapError.wrap(self._hide_parent_exception):
            self.conn.sasl_interactive_bind_s('', ldap.sasl.external())

    def bind_sasl_gssapi(self) -> None:  # pragma: no cover
        """Authenticate via GSSAPI e.g. via Kerberos ticket."""
        with errors.LdapError.wrap(self._hide_parent_exception):
            self.conn.sasl_interactive_bind_s('', ldap.sasl.gssapi())

    def bind_oauthbearer(self, authzid, token) -> None:  # pragma: no cover; requires SASL module
        """Authenticate via OAuth 2.0 Access Token."""
        oauth = ldap.sasl.sasl(
            {
                ldap.sasl.CB_AUTHNAME: authzid,
                ldap.sasl.CB_PASS: token,
            },
            'OAUTHBEARER',
        )
        with errors.LdapError.wrap(self._hide_parent_exception):
            self.conn.sasl_interactive_bind_s('', oauth)

    def _restore_options(self) -> None:
        for option, value in self._options:
            self.set_option(option, value, append=False)

    def _restore_auth_state(self) -> None:
        if self._last_auth_state:
            with errors.LdapError.wrap(self._hide_parent_exception):
                getattr(self.conn, self._last_auth_state[0])(*self._last_auth_state[1:])

    def unbind(self, *, controls: Controls | None = None) -> Result:
        """Unbind."""
        self._last_auth_state = None
        try:
            conn = self.conn
        except RuntimeError:  # not connected
            return None
        try:
            response = self._execute(conn, conn.unbind_ext, **Controls.expand(controls))
        except AttributeError as exc:  # duplicated unbind
            if exc.args and exc.args[0] == "ReconnectLDAPObject has no attribute '_l'":
                return None
            raise  # pragma: no cover; not possible
        return Result.from_response(None, None, controls, response)

    def whoami(self, *, controls: Controls | None = None) -> DN | str:
        """Get authenticated user DN (authzid). "Who am I?" Operation."""
        try:
            with errors.LdapError.wrap(self._hide_parent_exception):
                dn = self.conn.whoami_s(**Controls.expand(controls))
        except RuntimeError:
            return None
        if dn.startswith('dn:'):
            return DN(dn.removeprefix('dn:'))
        return dn  # pragma: no cover

    def change_password(self, dn: DN | str, old_password: str, new_password: str, *, controls: Controls | None = None) -> Result:
        """Change password."""
        conn = self.conn
        response = self._execute(conn, conn.passwd, str(dn), old_password, new_password, **Controls.expand(controls))
        return Result.from_response(dn, None, controls, response)  # pragma: no cover

    def exists(self, dn: DN | str, unique: bool = False, *, controls: Controls | None = None) -> bool:
        """Check if LDAP object exists."""
        try:
            self.get(dn, ['1.1'], unique=unique, controls=controls)
        except errors.NoSuchObject:
            return False
        return True

    def get(self, dn: DN | str, attrs=None, filter_expr='(objectClass=*)', *, unique: bool = False, controls: Controls | None = None) -> Result:
        """Get a LDAP object."""
        for obj in self.search(base=dn, scope=Scope.BASE, filter_expr=filter_expr, attrs=attrs, unique=unique, controls=controls):
            return obj
        return None  # pragma: no cover; impossible
        # obj, = [_ for _ in self.search_iter(base=dn, scope=Scope.BASE, filter_expr=filter_expr, attrs=attrs, unique=unique, controls=controls)]  # noqa: E501
        # return obj[0]
        # # GC calls gen.aclose() causing unnecessary .cancel() to be called:
        # # return next(self.search_iter(base=dn, scope=Scope.BASE, filter_expr=filter_expr, attrs=attrs, unique=unique, controls=controls))

    def get_attr(self, dn, attr, filter_expr='(objectClass=*)', *, unique: bool = False, controls: Controls | None = None) -> list[bytes]:
        """Get attribute of an LDAP object."""
        attributes = (self.get(dn, attrs=[attr], filter_expr=filter_expr, unique=unique, controls=controls)).attr
        try:
            return attributes[attr]
        except KeyError:
            self.get_schema()
            return attributes[attr]

    def search_iter(
        self,
        base: DN | str = '',
        scope: Scope = Scope.SUBTREE,
        filter_expr: str = '(objectClass=*)',
        attrs: list[str] | None = None,
        *,
        unique: bool = False,
        sizelimit: bool | None = None,
        sorting: Sorting | None = None,
        controls: Controls | None = None,
        _attrsonly: bool = False,
    ) -> AsyncGenerator[Result, None]:
        """Search iterative for DN and Attributes of LDAP objects."""
        conn = self.conn
        all_results = []
        if sorting:
            controls = Controls.set_server(controls, server_side_sorting(*sorting, criticality=True))
        # sizelimit = 1 if unique else sizelimit
        try:
            for response in self._execute_iter(
                conn,
                conn.search_ext,
                str(base),
                scope,
                filterstr=filter_expr,
                attrlist=attrs,
                attrsonly=int(_attrsonly),
                **Controls.expand(controls),
                timeout=self.timeout,
                sizelimit=sizelimit or OptionValue.NoLimit,
            ):
                Result.set_controls(response, controls)
                results = [Result.from_response(dn, attributes, controls, response) for dn, attributes in response.data]
                all_results.extend(results)
                if unique and len(all_results) > 1:
                    raise errors.NotUnique(all_results)
                try:
                    yield from results
                except GeneratorExit as exc:
                    with contextlib.suppress(errors.NoSuchOperation):
                        # self.cancel(response.msgid)  # better do it immediately
                        self.cancel(response.msgid)
                    raise exc from exc
        except errors.NoSuchObject as no_object_error:
            no_object_error.base_dn = DN.get(base)
            no_object_error.filter = filter_expr
            no_object_error.scope = scope
            no_object_error.attrs = attrs
            raise
        # except errors.SizelimitExceeded:
        #     if not unique:
        #         raise
        #     raise errors.NotUnique() from None

    def search(
        self,
        base: DN | str = '',
        scope: Scope = Scope.SUBTREE,
        filter_expr: str = '(objectClass=*)',
        attrs: list[str] | None = None,
        *,
        unique: bool = False,
        sizelimit: bool | None = None,
        sorting: Sorting | None = None,
        controls: Controls | None = None,
        _attrsonly: bool = False,
    ) -> AsyncGenerator[Result, None]:
        """Search for DN and Attributes of LDAP objects."""
        conn = self.conn
        all_results = []
        if sorting:
            controls = Controls.set_server(controls, server_side_sorting(*sorting))
        try:
            response = self._execute(
                conn,
                conn.search_ext,
                str(base),
                scope,
                filterstr=filter_expr,
                attrlist=attrs,
                attrsonly=int(_attrsonly),
                **Controls.expand(controls),
                timeout=self.timeout,
                sizelimit=sizelimit or OptionValue.NoLimit,
            )
            Result.set_controls(response, controls)
            results = [Result.from_response(dn, attributes, controls, response) for dn, attributes in response.data]
            all_results.extend(results)
            if unique and len(all_results) > 1:
                raise errors.NotUnique(all_results)
        except errors.NoSuchObject as no_object_error:
            no_object_error.base_dn = DN.get(base)
            no_object_error.filter = filter_expr
            no_object_error.scope = scope
            no_object_error.attrs = attrs
            raise
        return results

    def search_dn(
        self,
        base: DN | str = '',
        scope: Scope = Scope.SUBTREE,
        filter_expr: str = '(objectClass=*)',
        *,
        unique: bool = False,
        sizelimit: bool | None = None,
        sorting: Sorting | None = None,
        controls: Controls | None = None,
    ) -> AsyncGenerator[DN, None]:
        """Search for DNs of LDAP objects."""
        # FIXME: the following hangs forever as the iterative search is unfinished while the FD reader is replaced
        # for result in self.search(
        #     base, scope, filter_expr, ['1.1'], unique=unique, sizelimit=sizelimit, sorting=sorting, controls=controls, _attrsonly=True
        # ):
        for result in self.search(
            base, scope, filter_expr, [], unique=unique, sizelimit=sizelimit, sorting=sorting, controls=controls, _attrsonly=True
        ):
            yield result.dn

    def search_paginated(
        self,
        base: DN | str = '',
        scope: Scope = Scope.SUBTREE,
        filter_expr: str = '(objectClass=*)',
        attrs: list[str] | None = None,
        *,
        page_size: int = 100,
        sorting: Sorting,
        unique: bool = False,
        sizelimit: bool | None = None,
        controls: Controls | None = None,
    ) -> AsyncGenerator[Result, None]:
        """Search paginated using Virtual List View control."""
        controls = Controls.set_server(controls, server_side_sorting(*sorting))

        res_vlv = virtual_list_view.response()
        context_id = None
        length = None
        last_page = None
        page = 1
        while True:
            offset = ((page or 1) - 1) * page_size
            pagination = virtual_list_view(
                before_count=0,
                after_count=page_size - 1,
                offset=offset + 1,
                content_count=0,
                greater_than_or_equal=None,
                context_id=context_id,
                criticality=True,
            )
            controls = Controls.set_server(controls, pagination)
            if length is not None and offset > length:
                break  # end reached

            vlv = None
            current = None
            for entry_number, result in enumerate(
                self.search(base, scope, filter_expr, attrs, unique=unique, sizelimit=sizelimit, controls=controls), 1
            ):
                if last_page is None:
                    vlv = controls.get(res_vlv)
                    length = vlv.contentCount
                    last_page = math.ceil(length / (page_size or length))
                result.page = Page(
                    page=page,
                    entry=entry_number,
                    page_size=page_size,
                    results=length,
                    last_page=last_page,
                )
                current = result
                yield result

            if current is None:  # no search results
                break

            page += 1
            vlv = controls.get(res_vlv)
            context_id = vlv.context_id
            length = vlv.contentCount

    def search_paged(
        self,
        base: DN | str = '',
        scope: Scope = Scope.SUBTREE,
        filter_expr: str = '(objectClass=*)',
        attrs: list[str] | None = None,
        page_size: int = 100,
        *,
        unique: bool = False,
        sizelimit: bool | None = None,
        sorting: Sorting | None = None,
        controls: Controls | None = None,
    ) -> AsyncGenerator[Result, None]:
        """Search paginated using SimplePagedResults control."""
        pagination = simple_paged_results(size=page_size, cookie='', criticality=True)
        controls = Controls.append_server(controls, pagination)
        if sorting:
            controls = Controls.set_server(controls, server_side_sorting(*sorting))
        page = 0
        while True:
            current = None
            page += 1
            entry_number = 0
            for result in self.search_iter(base, scope, filter_expr, attrs, unique=unique, sizelimit=sizelimit, controls=controls):
                entry_number += 1  # noqa: SIM113
                result.page = Page(page=page, entry=entry_number, page_size=page_size)
                current = result
                yield result

            if current is None:  # no search results
                break
            control = controls.get(pagination)
            if not control:  # pragma: no cover
                break  # Server doesn't support pagination

            pagination.cookie = controls.get(pagination).cookie
            if not pagination.cookie:
                break

    def add(
        self,
        dn: DN | str,
        attrs: dict | Attributes,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Create a LDAP object."""
        al = ldap.modlist.addModlist(attrs)
        return self.add_al(dn, al, controls=controls)

    def add_al(
        self,
        dn: DN | str,
        al: LDAPAddList,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Create a LDAP object from addlist."""
        conn = self.conn
        response = self._execute(conn, conn.add_ext, str(dn), al, **Controls.expand(controls))
        return Result.from_response(dn, None, controls, response)

    def modify(
        self,
        dn: DN | str,
        oldattr: dict | Attributes,
        newattr: dict | Attributes,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Modify a LDAP object."""
        ml = ldap.modlist.modifyModlist(oldattr, newattr)
        return self.modify_ml(dn, ml, controls=controls)

    def modify_ml(
        self,
        dn: DN | str,
        ml: LDAPModList,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Modify a LDAP object from modlist."""
        conn = self.conn
        new_dn = self._compute_changed_dn(DN(dn), ml)
        if dn != new_dn:
            dn = (self.rename(dn, new_dn)).dn
        response = self._execute(conn, conn.modify_ext, str(dn), ml, **Controls.expand(controls))
        return Result.from_response(dn, None, controls, response)

    @classmethod
    def _compute_changed_dn(cls, dn: DN, ml: list[tuple[int, str, list[bytes] | None]]) -> DN:
        """
        Get changed DN.

        >>> Connection._compute_changed_dn('cn=foo,dc=bar', [(ldap.MOD_REPLACE, 'cn', b'foo')])
        'cn=foo,dc=bar'
        >>> Connection._compute_changed_dn('cn=foo,dc=bar', [(ldap.MOD_REPLACE, 'cn', b'bar')])
        'cn=bar,dc=bar'
        >>> Connection._compute_changed_dn('cn=foo,dc=bar', [(ldap.MOD_REPLACE, 'cn', b'föo')]) == 'cn=föo,dc=bar'
        True
        """
        rdn = dn.rdns[0]
        dn_vals = {x[0].lower(): x[1] for x in rdn}
        new_vals = {
            key.lower(): val.decode('UTF-8') if isinstance(val, bytes) else val[0].decode('UTF-8')
            for op, key, val in ml
            if key.lower() in dn_vals and val and op != ldap.MOD_DELETE
        }
        new_rdn_ava = [(x, new_vals.get(x.lower(), dn_vals[x.lower()]), ldap.AVA_STRING) for x in [y[0] for y in rdn]]
        new_rdn = DN(
            ldap.dn.dn2str(
                [
                    [(key, val, ava_type) for key, val, ava_type in new_rdn_ava],
                ],
            ),
        )
        if dn[0] != new_rdn:
            return new_rdn + dn.parent
        return dn

    def move(
        self,
        dn: DN | str,
        newposition: DN | str,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Move a LDAP object."""
        dn = DN.get(dn)
        newposition = DN.get(newposition)
        return self.rename(dn, dn[0] + newposition, delete_old=True, controls=controls)

    def rename(
        self,
        dn: DN | str,
        newdn: DN | str,
        delete_old: bool = True,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Rename a LDAP object."""
        conn = self.conn
        newdn = DN.get(newdn)
        response = self._execute(conn, conn.rename, str(dn), str(newdn[0]), str(newdn.parent), int(delete_old), **Controls.expand(controls))
        return Result.from_response(newdn, None, controls, response)

    def modrdn(
        self,
        dn: DN | str,
        newrdn: DN | str,
        delete_old: bool = True,
        *,
        controls: Controls | None = None,
    ) -> Result:
        """Rename a LDAP object."""
        return self.rename(dn, DN.get(newrdn) + DN.get(dn).parent, delete_old, controls=controls)

    def delete(self, dn: DN | str, *, controls: Controls | None = None) -> Result:
        """Delete a LDAP object."""
        conn = self.conn
        response = self._execute(conn, conn.delete_ext, str(dn), **Controls.expand(controls))
        return Result.from_response(dn, None, controls, response)

    def delete_recursive(self, dn: DN | str, *, controls: Controls | None = None) -> Result:
        """Delete a LDAP object recursively."""
        try:
            return self.delete(dn, controls=controls)
        except errors.NotAllowedOnNonleaf:
            for child in self.search_dn(dn, Scope.ONELEVEL):
                self.delete_recursive(child)
        return self.delete(dn, controls=controls)

    def compare(
        self,
        dn: DN | str,
        attr: str,
        value: bytes,
        *,
        controls: Controls | None = None,
    ) -> bool:
        """Compare the value of an LDAP object."""
        conn = self.conn
        try:
            self._execute(conn, conn.compare_ext, str(dn), attr, value, **Controls.expand(controls))
        except errors.NoSuchObject as no_object_error:
            no_object_error.base_dn = DN.get(dn)
            raise
        except errors.CompareTrue:
            return True
        except errors.CompareFalse:
            return False
        raise RuntimeError()  # pragma: no cover; impossible

    def compare_dn(self, entry: DN | str, dn: DN | str) -> bool:
        """Compare LDAP DN with existing entry."""
        dn = DN.get(dn)
        entry = DN.get(entry)

        for i, parent in enumerate(entry.walk()):
            for attr, value, _ in dn.rdns[-i - 1]:
                try:
                    equal = self.compare(str(parent), attr, value)
                    if not equal:  # pragma: no cover; https://github.com/nedbat/coveragepy/issues/2014
                        return False
                except errors.NoSuchObject:
                    if attr == entry.rdns[-1][0][0]:
                        continue
                    raise
        return True

    def get_root_dse(self, attrs: list[str] | None = None, filter_expr: str = '(objectClass=*)') -> Result:
        """Get Root DSE (Directory Server Entry)."""
        return self.get('', attrs or ['*', '+'], filter_expr=filter_expr)

    def get_naming_contexts(self) -> list[str]:
        """Return namingContexts of Root DSE."""
        result = self.get_attr('', 'namingContexts')
        return [x.decode('UTF-8') for x in result]

    def abandon(self, msgid: int, *, controls: Controls | None = None) -> Result:
        """Abandon a LDAP operation."""
        log.debug('Abandon: %s', msgid)
        conn = self.conn
        response = self._execute(conn, conn.abandon_ext, msgid, **Controls.expand(controls))
        return Result.from_response(None, None, controls, response)

    def cancel(self, msgid: int, *, controls: Controls | None = None) -> bool:
        """Cancel a LDAP operation."""
        log.debug('Cancel: %s', msgid)
        try:
            conn = self.conn
        except RuntimeError:
            return False
        try:
            self._execute(conn, conn.cancel, msgid, **Controls.expand(controls))
        except errors.NoSuchOperation:
            log.warning('Cancel failed', extra={'msgid': msgid})
            raise
        except (errors.Cancelled, errors.Success):  # pragma: no cover; theoretically, according to python-ldap
            return True
        except errors.TooLate:  # pragma: no cover
            return False
        else:  # pragma: no cover
            return True

    def refresh_ttl(self, dn: DN | str, ttl: int):
        """Perform Refresh extended operation."""
        from ldap.extop.dds import RefreshRequest, RefreshResponse  # noqa: PLC0415

        req = RefreshRequest(RefreshRequest.requestName, str(dn), ttl)
        req.requestValue = b''  # req.encodedRequestValue()
        self.extended(req, RefreshResponse)

    def extended(self, request: ldap.extop.ExtendedRequest, response_class=None, *, controls: Controls | None = None) -> Result | Any:
        """Perform extended operation."""
        conn = self.conn
        response = self._execute(conn, conn.extop, request, **Controls.expand(controls))

        if response_class:
            with errors.LdapError.wrap(self._hide_parent_exception):
                if response_class.responseName == response.name:  # pragma: no cover
                    return response_class(response.name, response.value)
                raise errors.ProtocolError({'desc': 'OID in extended response does not match response class.'})
        return Result.from_response(None, None, controls, response)  # pragma: no cover

    def __getstate__(self) -> dict:
        """Return state for pickle."""
        return {slot: getattr(self, slot) for slot in set(self.__slots__) - {'_conn'} | {'connected'} if not slot.startswith('__')}

    def __setstate__(self, state: dict) -> None:
        """Set state for pickle."""
        self._conn = None
        connected = state.pop('connected', None)
        for slot, value in state.items():
            setattr(self, slot, value)
        if connected:
            self.connect()
            self._restore_options()
        self._restore_auth_state()

    def _execute(self, conn: LDAPObject, operation: Callable, *args: Any, **kwargs: Any) -> _Response:
        """Execute the operation and wait asynchronously for the result."""
        msgid = self._retry(self.request, operation, *args, **kwargs)
        if msgid is None:  # abandon_ext, unbind_ext
            return _Response(None, None, msgid, [], None, None)
        response = None
        for resp in self._poll(conn, msgid, 1):
            if response is not None:  # pragma: no cover
                raise RuntimeError('Wrong method used! Use _execute_iter instead!')  # noqa: TRY003
            response = resp
        return response

    def _execute_iter(self, conn: LDAPObject, operation: Callable, *args: Any, **kwargs: Any) -> AsyncGenerator[_Response, None]:
        """Execute the operation and yield the results asynchronously."""
        msgid = self._retry(self.request, operation, *args, **kwargs)
        if msgid is None:  # abandon_ext, unbind_ext
            return
        yield from self._poll(conn, msgid, 0)

    def get_result(self, conn: LDAPObject, msgid: ResponseType = ResponseType.Any, _all: int = 0, timeout: int = 0) -> _Response:
        """Get the LDAP result for the given msgid."""
        log.debug('result(%r, timeout=%r)', msgid, timeout, extra={'MSGID': msgid, 'ALL': _all, 'TIMEOUT': timeout, 'FUNC': 'result'})
        try:
            with errors.LdapError.wrap(self._hide_parent_exception):
                response = _Response(*conn.result4(msgid, all=_all, timeout=timeout))
        except (errors.LdapError, OSError) as exc:
            log.debug('result(%r) -> raised %r', msgid, exc, extra={'MSGID': msgid, 'OPERATION': 'result', 'EXCEPTION': str(exc)})
            raise
        log.debug('result(%r) -> %s', msgid, repr(response)[:200], extra={'MSGID': msgid, 'OPERATION': 'result'})
        return response

    def request(self, operation: Callable, *args: Any, **kwargs: Any) -> int | None:
        """Make the LDAP request for the given operation."""
        op = operation.__name__
        arg_str = ', '.join(map(repr, args)) if 'bind' not in op else ''
        kw = ', '.join(f'{k}={v!r}' for k, v in kwargs.items()) if 'bind' not in op else ''
        log.debug('Request %s(%s%s%s)', op, arg_str, ', ' if kw else '', kw, extra={'OPERATION': op, 'ARGUMENTS': arg_str, 'KEYWORDS': kw})
        try:
            with errors.LdapError.wrap(self._hide_parent_exception):
                msgid = operation(*args, **kwargs)
        except (errors.LdapError, OSError) as exc:
            log.debug('%s() -> %r', op, exc, extra={'OPERATION': op, 'EXCEPTION': str(exc)})
            raise
        log.debug('%s() -> %r', op, msgid, extra={'OPERATION': op, 'MSGID': msgid})
        return msgid

    def _retry(self, func: Callable, *args: Any, **kwargs: Any) -> Any:
        """Retry operation or reconnect if necessary."""
        max_attempts = attempts = self.max_connection_attempts
        while attempts:
            try:
                if attempts != max_attempts:
                    # do a sync reconnect, make sure we don't end in recursion
                    self.reconnect()  # FIXME: recursion!?
                    # self.conn.reconnect(self.uri)
                    # self._restore_auth_state()  # FIXME: recursion

                return func(*args, **kwargs)
            except (errors.ServerDown, errors.Unavailable, errors.ConnectError, errors.Timeout):
                attempts -= 1
                if not attempts:
                    raise
                time.sleep(self.retry_delay)
        raise RuntimeError()  # pragma: no cover; impossible

    def _poll(self, conn: LDAPObject, msgid: ResponseType = ResponseType.Any, _all: int = 0) -> Generator[_Response, None]:  # pragma: no cover
        """Wait synchronously for operation to succeed."""
        # this method must only used by the synchronous variant of this class
        while True:
            try:
                response = self._retry(self.get_result, conn, msgid, _all=_all, timeout=self.timeout)
            except errors.NoResultsReturned:  # pragma: no cover
                break

            rtype = response.type
            if rtype is None:
                continue

            yield response
            if rtype == ldap.RES_SEARCH_ENTRY:
                continue
            if rtype == ldap.RES_SEARCH_RESULT:
                break
            break
